from django.db.models import QuerySet

from wbcore import viewsets
from wbcore.contrib.example_app.filters import LeagueFilter, LeagueSportFilter
from wbcore.contrib.example_app.models import League
from wbcore.contrib.example_app.serializers import (
    LeagueModelSerializer,
    LeagueRepresentationSerializer,
)
from wbcore.contrib.example_app.viewsets.displays import (
    LeagueDisplayConfig,
    LeagueSportDisplayConfig,
)
from wbcore.contrib.example_app.viewsets.endpoints import LeagueSportEndpointConfig
from wbcore.contrib.example_app.viewsets.titles import (
    LeagueSportTitleConfig,
    LeagueTitleConfig,
)


class LeagueModelViewSet(viewsets.ModelViewSet):
    display_config_class = LeagueDisplayConfig
    queryset = League.objects.all()
    search_fields = ("name", "sport__name")
    serializer_class = LeagueModelSerializer
    title_config_class = LeagueTitleConfig
    ordering_fields = (
        "name",
        "country__name",
        "established_date",
        "commissioner__computed_str",
        "website",
        "sport__name",
        "points_per_win",
        "points_per_draw",
        "points_per_loss",
    )
    ordering = ("name", "sport__name")
    filterset_class = LeagueFilter

    def get_queryset(self) -> QuerySet[League]:
        return super().get_queryset().select_related("sport", "country", "commissioner")


class LeagueSportModelViewSet(LeagueModelViewSet):
    display_config_class = LeagueSportDisplayConfig
    ordering_fields = (
        "name",
        "country__name",
        "established_date",
        "commissioner__computed_str",
        "website",
        "points_per_win",
        "points_per_draw",
        "points_per_loss",
    )
    ordering = ("name",)
    filterset_class = LeagueSportFilter
    endpoint_config_class = LeagueSportEndpointConfig
    title_config_class = LeagueSportTitleConfig

    def get_queryset(self) -> QuerySet[League]:
        queryset = super().get_queryset()
        if sport_id := self.kwargs.get("sport_id"):
            queryset = queryset.filter(sport=sport_id)
        return queryset


class LeagueRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = League.objects.all()
    serializer_class = LeagueRepresentationSerializer
    search_fields = ("name",)
    filterset_fields = ("sport",)



import factory

from .models import EntityTest


class EntityTestFactory(factory.django.DjangoModelFactory):
    dl_parameters = {
        "data": {"path": "wbcore.contrib.dataloader.tests.test.dataloaders.dataloaders.RandomData", "parameters": {}}
    }
    name = factory.Faker("pystr")

    class Meta:
        model = EntityTest



import pytest

from django.apps import apps
from django.db import connection
from django.db.models.signals import pre_migrate
from pytest_factoryboy import register
from rest_framework.test import APIRequestFactory, APIClient
from wbcore.contrib.authentication.factories import UserFactory

from ..factories import ContinentFactory
from .signals import app_pre_migration

register(ContinentFactory)


pre_migrate.connect(app_pre_migration, sender=apps.get_app_config("geography"))


@pytest.fixture()
def request_factory():
    return APIRequestFactory()


@pytest.fixture()
def super_user():
    return UserFactory(is_superuser=True)



from pytest_factoryboy import register
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.directory.factories import EntryFactory, PersonFactory
from wbcore.tests.conftest import *

from ..factories import BuildingFactory, CalendarItemFactory, ConferenceRoomFactory
from .signals import *

register(CalendarItemFactory)
register(BuildingFactory)
register(EntryFactory)
register(PersonFactory)
register(UserFactory)
register(ConferenceRoomFactory)



from rest_framework import serializers

from .mixins import WBCoreSerializerFieldMixin
from .types import WBCoreType


class ImageField(WBCoreSerializerFieldMixin, serializers.ImageField):
    field_type = WBCoreType.IMAGE.value


class UnsafeImageField(WBCoreSerializerFieldMixin, serializers.FileField):
    field_type = WBCoreType.IMAGE.value


class ImageURLField(WBCoreSerializerFieldMixin, serializers.URLField):
    field_type = WBCoreType.IMAGE.value


class FileField(WBCoreSerializerFieldMixin, serializers.FileField):
    field_type = WBCoreType.FILE.value



from django.contrib import admin

from .models import Asset


@admin.register(Asset)
class AssetModelAdmin(admin.ModelAdmin):
    list_display = ("id", "file", "content_type")
    readonly_fields = ("content_type", "file_url_name")



import pytest
from django.apps import apps
from django.db import connection
from django.db.models.signals import pre_migrate
from pytest_factoryboy import register
from wbcore.contrib.documents.factories import (
    DocumentFactory,
    DocumentTypeFactory,
    ShareableLinkAccessFactory,
    ShareableLinkFactory,
)
from wbcore.contrib.geography.tests.signals import app_pre_migration
from wbcore.tests.conftest import *

register(DocumentFactory)
register(DocumentTypeFactory)
register(ShareableLinkAccessFactory)
register(ShareableLinkFactory)


@pytest.fixture(
    autouse=True, scope="session"
)  # Might want to find a way to registered default conftest logic automatically
def django_test_environment(django_test_environment):
    from django.apps import apps

    get_models = apps.get_models

    for m in [m for m in get_models() if not m._meta.managed]:
        m._meta.managed = True



from django_celery_beat.models import IntervalSchedule, PeriodicTask
from django.db import DEFAULT_DB_ALIAS
from django.apps import apps as global_apps


def initialize_task(app_config, verbosity=2, interactive=True, using=DEFAULT_DB_ALIAS, apps=global_apps, **kwargs):
    PeriodicTask.objects.update_or_create(
        task="wbcore.contrib.authentication.tasks.delete_unregistered_user_account",
        defaults={
            "name": "Authentication: Delete ghost user accounts",
            "interval": IntervalSchedule.objects.get_or_create(every=1, period=IntervalSchedule.DAYS)[0],
            "crontab": None,
        },
    )



from datetime import datetime
from decimal import Decimal
from typing import Any, Dict, Optional

from django.db import models

from wbcore.contrib.io.imports import ImportExportHandler

from .currency import CurrencyImportHandler


class CurrencyFXRatesImportHandler(ImportExportHandler):
    MODEL_APP_LABEL: str = "currency.CurrencyFXRates"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.currency_handler = CurrencyImportHandler(self.import_source)

    def _deserialize(self, data: Dict[str, Any]):
        data["date"] = datetime.strptime(data["date"], "%Y-%m-%d").date()
        data["currency"] = self.currency_handler.process_object(data["currency"])[0]
        data["value"] = round(Decimal(data["value"]), 6)

    def _get_instance(self, data: Dict[str, Any], history: Optional[models.QuerySet] = None, **kwargs) -> models.Model:
        return self.model.objects.filter(currency=data["currency"], date=data["date"]).first()

    def _create_instance(self, data: Dict[str, Any], **kwargs) -> models.Model:
        self.import_source.log += "\nCreate Fx Rate."
        return self.model.objects.create(**data, import_source=self.import_source)



# Generated by Django 5.0.12 on 2025-02-28 10:21

import modeltrans.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('directory', '0010_remove_addresscontact_city'),
    ]

    operations = [
        migrations.AddField(
            model_name='person',
            name='description',
            field=models.TextField(blank=True, default=''),
        ),
        migrations.AddField(
            model_name='person',
            name='i18n',
            field=modeltrans.fields.TranslationField(fields=['description'], required_languages=(), virtual_fields=True),
        ),
    ]



import pytest
from django.conf import settings
from django.db import models
from django.urls import get_resolver

from wbcore import serializers, viewsets
from wbcore.pandas.views import PandasAPIViewSet

from .mixins import (
    TestChartViewSet,
    TestModel,
    TestPandasView,
    TestRepresentationViewSet,
    TestSerializer,
    TestViewSet,
)
from .utils import get_all_subclasses, is_intermediate_table_m2m

""""
If CELERY_TASK_ALWAYS_EAGER is True, all tasks will be executed locally by blocking until the task returns.
apply_async() and Task.delay() will return an EagerResult instance, that emulates the API and behavior of AsyncResult, except the result is already evaluated.
That is, tasks will be executed locally instead of being sent to the queue.
This is useful mainly when running tests, or running locally without Celery workers.
"""
settings.CELERY_TASK_ALWAYS_EAGER = True

get_resolver().url_patterns


def modules_condition(module):
    return not module.startswith(("wbcore", "django", "rest_framework", "dynamic_preferences", "eventtools")) or (
        module.startswith("wbcore") and module.startswith("wbcore.contrib")
    )


default_config = {
    "models": list(
        filter(
            lambda x: modules_condition(x.__module__) and not x._meta.abstract and not is_intermediate_table_m2m(x),
            get_all_subclasses(models.Model),
        )
    ),
    "serializers": list(
        filter(lambda x: modules_condition(x.__module__), get_all_subclasses(serializers.ModelSerializer))
    ),
    "representations": list(
        filter(lambda x: modules_condition(x.__module__), get_all_subclasses(viewsets.RepresentationViewSet))
    ),
    "viewsets": list(
        filter(
            lambda x: modules_condition(x.__module__) and x not in get_all_subclasses(viewsets.InfiniteDataModelView),
            get_all_subclasses(viewsets.ModelViewSet),
        )
    ),
    "pandasviews": list(
        filter(
            lambda x: modules_condition(x.__module__),
            get_all_subclasses(PandasAPIViewSet),
        )
    ),
    "chartviewsets": list(
        filter(
            lambda x: modules_condition(x.__module__),
            get_all_subclasses(viewsets.ChartViewSet),
        )
    ),
}


class GenerateTest:
    def __init__(self, config):
        self.config = config

    def test_models(self, _model):
        my_test = TestModel(_model)
        my_test.execute_test()

    def test_serializers(self, _serializer):
        my_test = TestSerializer(_serializer)
        my_test.execute_test()

    def test_representationviewsets(self, rvs):
        my_test = TestRepresentationViewSet(rvs)
        my_test.execute_test()

    def test_modelviewsets(self, mvs, client):
        my_test = TestViewSet(mvs)
        my_test.execute_test_list_endpoint(client)
        my_test.execute_test_detail_endpoint()

    def test_pandasviews(self, pvs):
        my_test = TestPandasView(pvs)
        if my_test.model:
            my_test.execute_test()

    def test_chartviewsets(self, cvs):
        my_test = TestChartViewSet(cvs)
        my_test.execute_test()

    def __call__(self, test_class):
        @pytest.mark.parametrize("_model", self.config.get("models", []))
        def _test_models(_self, _model):
            self.test_models(_model)

        @pytest.mark.parametrize("_serializer", self.config.get("serializers", []))
        def _test_serializers(_self, _serializer):
            self.test_serializers(_serializer)

        @pytest.mark.parametrize("rvs", self.config.get("representations", []))
        def _test_representationviewsets(_self, rvs):
            self.test_representationviewsets(rvs)

        @pytest.mark.parametrize("mvs", self.config.get("viewsets", []))
        def _test_modelviewsets(_self, mvs, client):
            self.test_modelviewsets(mvs, client)

        @pytest.mark.parametrize("pvs", self.config.get("pandasviews", []))
        def _test_pandasviews(_self, pvs):
            self.test_pandasviews(pvs)

        @pytest.mark.parametrize("cvs", self.config.get("chartviewsets", []))
        def _test_chartviewsets(_self, cvs):
            self.test_chartviewsets(cvs)

        setattr(test_class, "test_models", _test_models)
        setattr(test_class, "test_serializers", _test_serializers)
        setattr(test_class, "test_representationviewsets", _test_representationviewsets)
        setattr(test_class, "test_modelviewsets", _test_modelviewsets)
        setattr(test_class, "test_pandasviews", _test_pandasviews)
        setattr(test_class, "test_chartviewsets", _test_chartviewsets)

        return test_class



import csv
import re
from _csv import Error

import numpy as np
import pandas as pd


def parse(import_source):
    # try to guess the delimiter
    try:
        with import_source.file.open(mode="r") as file:
            sep = csv.Sniffer().sniff(file.readline()).delimiter
    except Error:
        sep = ","

    df = pd.read_csv(import_source.file, sep=sep)
    df = df.loc[:, ~df.columns.str.contains("^Unnamed")]  # remove columns without names
    # in case we have import kwargs, we set them as column values
    columns_mapping = import_source.resource_kwargs.get("columns_mapping", {})
    extra_columns_values = import_source.resource_kwargs.get("extra_columns_values", {})
    for k, v in extra_columns_values.items():
        df[k] = v
    df = df.rename(columns=columns_mapping)
    df = df.replace([np.inf, -np.inf, np.nan], None)
    rows = []
    for row in df.convert_dtypes().to_dict(orient="records"):
        for k, v in row.items():
            if isinstance(v, str):
                # we remove any whitespace special characters
                row[k] = re.sub(r"\s+", "", v.strip())
        rows.append(row)
    data = {"data": rows}
    if extra_columns_values:
        data["history"] = extra_columns_values
    return data



from .base import WBCoreViewConfig


class PrimaryKeyViewConfig(WBCoreViewConfig):
    metadata_key = "pk"
    config_class_attribute = "primary_key_config_class"

    def get_metadata(self) -> int | str | None:
        return self.view.kwargs.get("pk", None)



from django.contrib.contenttypes.models import ContentType
from rest_framework.reverse import reverse

from wbcore import serializers
from wbcore.content_type.serializers import ContentTypeRepresentationSerializer
from wbcore.content_type.utils import get_view_content_type_id

from .models import Tag, TagGroup


class TagGroupRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="wbcore:tags:taggroup-detail")

    class Meta:
        model = TagGroup
        fields = ("id", "title", "_detail")


class TagGroupModelSerializer(serializers.ModelSerializer):
    @serializers.register_resource()
    def additional_resources(self, instance, request, user):
        return {
            "tags": reverse(
                "wbcore:tags:group-tag-list",
                args=[instance.id],
                request=request,
            )
        }

    class Meta:
        model = TagGroup
        read_only_fields = ("slug",)
        fields = ("id", "title", "_additional_resources")


class TagRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="wbcore:tags:tag-detail")
    _detail_preview = serializers.HyperlinkField(reverse_name="wbcore:tags:tag-detail")

    def get_filter_params(self, request):
        content_type_id = get_view_content_type_id(request.parser_context["view"])
        return {"content_type": content_type_id}

    class Meta:
        model = Tag
        read_only_fields = ("slug",)
        fields = (
            "id",
            "title",
            "computed_str",
            "color",
            "slug",
            "description",
            "_detail",
            "_detail_preview",
        )


class TagModelSerializer(serializers.ModelSerializer):
    _groups = TagGroupRepresentationSerializer(source="groups", many=True)
    _content_type = ContentTypeRepresentationSerializer(source="content_type")
    content_type = serializers.PrimaryKeyRelatedField(
        queryset=ContentType.objects.all(), required=False, allow_null=True
    )
    groups = serializers.PrimaryKeyRelatedField(
        queryset=TagGroup.objects.all(),
        many=True,
        default=serializers.DefaultFromKwargs("group_id"),
    )

    class Meta:
        model = Tag
        read_only_fields = ("slug",)
        fields = (
            "id",
            "title",
            "computed_str",
            "color",
            "content_type",
            "_content_type",
            "slug",
            "description",
            "groups",
            "_groups",
        )


class TagSerializerMixin(serializers.Serializer):
    tags = serializers.PrimaryKeyRelatedField(queryset=Tag.objects.all(), label="Tags", many=True)
    _tags = TagRepresentationSerializer(source="tags", many=True)



import functools
import operator

import pandas as pd
import plotly.graph_objects as go


def get_hovertemplate_timeserie(is_percent=False, currency="$"):
    if is_percent:
        return "%{x|%Y-%m-%d}, <b>%{y:.4%}</b>"
    else:
        return f"%{{x|%Y-%m-%d}}, <b>%{{y:.4s}}{currency}</b>"


def get_default_timeserie_figure(fig=None, add_rangeslider=True, is_legend_horizontal=True):
    if not fig:
        fig = go.Figure()
    fig.update_layout(
        margin=dict(l=10, r=10, t=10, b=10),
        # paper_bgcolor='rgba(0,0,0,0)',
        plot_bgcolor="rgba(0,0,0,0)",
        yaxis=dict(showline=True, linewidth=1, linecolor="black"),
        xaxis=dict(
            linewidth=0.5,
            showgrid=True,
            gridcolor="lightgray",
            gridwidth=1,
            type="date",
            # showspikes=True,
            # spikethickness=1,
            zeroline=True,
            showline=True,
            linecolor="grey",
            ticks="outside",
            tickwidth=1,
            tickcolor="black",
            ticklen=10,
            tickangle=45,
            tickformatstops=[
                go.layout.xaxis.Tickformatstop(dtickrange=[86400000, 604800000], value="%e %b %y"),
                go.layout.xaxis.Tickformatstop(dtickrange=[604800000, "M1"], value="%e %b %y"),
                go.layout.xaxis.Tickformatstop(dtickrange=["M1", "M12"], value="%b %y"),
                go.layout.xaxis.Tickformatstop(dtickrange=["M12", None], value="%Y"),
            ],
            rangeselector=dict(
                buttons=list(
                    [
                        dict(count=1, label="1m", step="month", stepmode="backward"),
                        dict(count=6, label="6m", step="month", stepmode="backward"),
                        dict(count=1, label="1y", step="year", stepmode="backward"),
                        dict(count=1, label="YTD", step="year", stepmode="todate"),
                        dict(step="all"),
                    ]
                ),
            ),
        ),
        autosize=True,
        hovermode="x unified",
    )
    fig.update_layout(hovermode="closest", hoverlabel_namelength=-1)
    fig.update_xaxes(showspikes=True)
    fig.update_yaxes(showspikes=True)
    if add_rangeslider:
        fig.update_layout(xaxis_rangeslider_visible=True)
    if is_legend_horizontal:
        fig.update_layout(legend_orientation="h")
    return fig


def get_timeseries_chart(x_data, y_data, red, green, blue):
    fig = go.Figure(
        [
            go.Scatter(
                x=x_data,
                y=y_data,
                fill="tozeroy",
                line=dict(color=f"rgb({red}, {green}, {blue})", width=1),
                fillcolor=f"rgba({red}, {green}, {blue}, 0.1)",
            )
        ]
    )
    fig.update_layout(
        paper_bgcolor="rgba(0,0,0,0)",
        plot_bgcolor="rgba(0,0,0,0)",
        yaxis=dict(
            title="",
            titlefont=dict(color="#000000"),
            tickfont=dict(color="#000000"),
            anchor="x",
            side="right",
            showline=True,
            linewidth=1,
            linecolor="black",
        ),
        yaxis_type="log",
        xaxis=dict(
            title="",
            titlefont=dict(color="#000000"),
            tickfont=dict(color="#000000"),
            showline=True,
            linewidth=0.5,
            linecolor="black",
            showgrid=True,
            gridcolor="lightgray",
            gridwidth=1,
        ),
        autosize=True,
        xaxis_rangeslider_visible=True,
    )

    return fig


def get_factsheet_timeseries_chart(df, color="rgb(120,214,255)", fillcolor="rgba(120,214,255, 0.5)", log_scale=True):
    df = df.sort_index()
    df.index = pd.to_datetime(df.index)
    trace = go.Scatter(x=df.index, y=df.iloc[:, 0], fill="tonexty", line=dict(color=color, width=1))
    fig = go.Figure()
    fig.add_trace(trace)

    fig.update_layout(
        margin=dict(l=0, r=0, t=0, b=0),
        paper_bgcolor="rgba(0,0,0,0)",
        plot_bgcolor="rgba(0,0,0,0)",
        xaxis=dict(
            showgrid=True,
            zeroline=True,
            gridwidth=1,
            showline=True,
            gridcolor="grey",
            linewidth=1,
            linecolor="grey",
            ticks="outside",
            tickwidth=1,
            tickcolor="black",
            ticklen=10,
            tickangle=45,
            tickformatstops=[
                go.layout.xaxis.Tickformatstop(dtickrange=[86400000, 604800000], value="%e %b %y"),
                go.layout.xaxis.Tickformatstop(dtickrange=[604800000, "M1"], value="%e %b %y"),
                go.layout.xaxis.Tickformatstop(dtickrange=["M1", "M12"], value="%b %y"),
                go.layout.xaxis.Tickformatstop(dtickrange=["M12", None], value="%Y"),
            ],
            rangeselector=dict(
                buttons=list(
                    [
                        dict(count=1, label="1m", step="month", stepmode="backward"),
                        dict(count=6, label="6m", step="month", stepmode="backward"),
                        dict(count=1, label="1y", step="year", stepmode="backward"),
                        dict(count=1, label="YTD", step="year", stepmode="todate"),
                        dict(step="all"),
                    ]
                ),
                x=0,
                xanchor="left",
                y=1.1,
                yanchor="top",
            ),
            rangeslider=dict(visible=True),
            type="date",
            title="",
            titlefont=dict(color="#000000"),
            tickfont=dict(color="#000000"),
        ),
        yaxis=dict(
            title="",
            titlefont=dict(color="#000000"),
            tickfont=dict(color="#000000"),
            anchor="x",
            side="right",
            autorange=True,
            fixedrange=False,
            showgrid=False,
            zeroline=False,
        ),
    )
    if log_scale:
        fig.update_layout(yaxis=dict(type="log"))
    return fig


def get_horizontal_barplot(
    df,
    x_label="weighting",
    y_label="aggregated_title",
    colors=["#B4DAFF"],
    colors_label=None,
    drop_null_x_value: bool = True,
):
    df = df.dropna(how="any")
    if len(df.shape) == 1:
        df = df.sort_values(ascending=True)
        df = df.reset_index()
        df.columns = [y_label, x_label]

    if not isinstance(x_label, list):
        x_label = [x_label]

    if drop_null_x_value:
        df = df[functools.reduce(operator.and_, [df[label] != 0 for label in x_label])]

    data = []
    df = df.sort_values(by=[x_label[0]], ascending=False)
    if colors_label:
        colors = df[colors_label]
    opacity = 1
    for i, label in enumerate(x_label):
        data.append(
            go.Bar(
                name=label,
                x=df.loc[:, label],
                y=df.loc[:, y_label],
                opacity=opacity,
                orientation="h",
                marker=dict(color=colors),
            )
        )
        opacity -= 1 / (len(x_label) + 1)

    fig = go.Figure(data=data)

    fig.update_layout(
        barmode="group",
        yaxis_autorange="reversed",
        xaxis=dict(showgrid=False, showline=False, zeroline=False, tickformat=".1%"),
        yaxis=dict(
            showgrid=False,
            showline=False,
            zeroline=False,
        ),
        margin=dict(b=0, r=20, l=20, t=0, pad=20),
        paper_bgcolor="rgba(0,0,0,0)",
        plot_bgcolor="rgba(0,0,0,0)",
        # autosize=False,
        font=dict(family="roboto", size=12, color="black"),
    )
    height = df.shape[0] * 40
    if height != 0:
        fig.update_layout(height=max(height, 100))

    return fig


def get_piechart(
    df,
    x_label="weighting",
    y_label="aggregated_title",
    hoverinfo="label+text",
    max_number_label=3,
    height=280,
    with_legend=True,
    default_normalize=True,
    colors_map=None,
    colors_label=None,
):
    df = df.sort_values(by=[x_label], ascending=False)
    values = df[x_label]
    label = df[y_label]
    if default_normalize:
        values = df.loc[(df[x_label] > 0), x_label]
        label = df.loc[(df[x_label] > 0), y_label]

    text = values.map("{:,.1%}".format)
    fig = go.Figure(data=[go.Pie(direction="clockwise", labels=label, values=values.abs(), text=text)])
    fig.update_traces(hoverinfo=hoverinfo, textinfo="none", textfont_size=8)
    if colors_map:
        fig.update_traces(marker=dict(colors=colors_map))
    elif colors_label:
        fig.update_traces(marker=dict(colors=df[colors_label]))
    fig.update_layout(
        margin=go.layout.Margin(b=0, l=0, r=0, t=0, pad=0),
        # showlegend=False,
        paper_bgcolor="rgba(0,0,0,0)",
        plot_bgcolor="rgba(0,0,0,0)",
    )
    if with_legend:
        fig.update_layout(
            height=height,
            legend=go.layout.Legend(
                x=0,
                y=-1,
                yanchor="bottom",
                xanchor="left",
                bgcolor="rgba(0,0,0,0)",
                font=dict(family="roboto", size=11, color="black"),
            ),
        )
    else:
        fig.update_layout(showlegend=False)
    return fig



from django.utils.translation import gettext_lazy as _
from rest_framework.reverse import reverse

from wbcore import serializers
from wbcore.contrib.documents.models import Document
from wbcore.contrib.documents.serializers import DocumentTypeRepresentationSerializer


class DocumentModelSerializer(serializers.ModelSerializer):
    _document_type = DocumentTypeRepresentationSerializer(source="document_type")
    system_created = serializers.BooleanField(read_only=True, label=_("System Created"))

    @serializers.register_resource()
    def shareable_links(self, instance, request, user):
        base_url = reverse("wbcore:documents:link-list", request=request)
        return {"shareable_links": f"{base_url}?document={instance.id}"}

    @serializers.register_resource()
    def download_file(self, instance, request, user):
        return {"download_file": instance.file.url}

    @serializers.register_resource()
    def send_mail(self, instance, request, user):
        return {"send_mail": reverse("wbcore:documents:document-sendmail", args=[instance.id], request=request)}

    @serializers.register_only_instance_resource()
    def relationships(self, instance, request, user, view):
        return {
            "relationships": reverse(
                "wbcore:documents:document-documentmodelrelationship-list", args=[instance.id], request=request
            )
        }

    class Meta:
        model = Document
        fields = (
            "id",
            "name",
            "description",
            "file",
            "document_type",
            "_document_type",
            "system_created",
            "system_key",
            "updated",
            "created",
            "valid_from",
            "valid_until",
            "_additional_resources",
        )


class ReadOnlyDocumentModelSerializer(DocumentModelSerializer):
    class Meta(DocumentModelSerializer.Meta):
        read_only_fields = DocumentModelSerializer.Meta.fields


class DocumentRepresentationSerializer(serializers.RepresentationSerializer):
    endpoint = "wbcore:documents:documentrepresentation-list"
    _detail = serializers.HyperlinkField(reverse_name="wbcore:documents:document-detail")
    _detail_preview = serializers.HyperlinkField(reverse_name="wbcore:documents:document-detail")

    class Meta:
        model = Document
        fields = ("id", "_detail", "_detail_preview", "name")



from .base import WBCoreViewConfig


class OrderingFieldsViewConfig(WBCoreViewConfig):
    metadata_key = "ordering_fields"
    config_class_attribute = "ordering_fields_config_class"

    def get_metadata(self) -> dict[str, str]:
        ordering_fields = dict()

        for ordering_field in self.view.get_ordering_fields():
            field_name = ordering_field.replace("__nulls_last", "").replace("__nulls_first", "")

            # TODO: Potentially refactor
            serializer = None
            if serializer_method := getattr(self.view, "get_serializer", None):
                serializer = serializer_method()

            if "__" in ordering_field and (not serializer or field_name not in serializer.fields.keys()):
                field_name = field_name.split("__")[0]

            ordering_fields[field_name] = ordering_field

        return ordering_fields



import pytest
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from faker import Faker
from wbcore.permissions.shortcuts import get_internal_users

User = get_user_model()

fake = Faker()


def create_internal_user():
    from wbcore.permissions.registry import user_registry

    user = User.objects.create(is_active=True, username=fake.user_name(), email=fake.email())
    permission = Permission.objects.get_or_create(
        content_type=ContentType.objects.get_for_model(User), codename="is_internal_user"
    )[0]
    user.user_permissions.add(permission)
    user_registry.reset_cache()
    return user


@pytest.mark.django_db
class TestBackend:
    def test_get_internal_users(self):
        user = create_internal_user()
        assert set(get_internal_users()) == {user}



from rest_framework.test import APIRequestFactory

from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.workflow.models import ProcessStep
from wbcore.contrib.workflow.viewsets.display.process import (
    get_state_formatting,
    get_state_legend,
)


class TestProcessStep:
    api_factory = APIRequestFactory()

    def test_get_state_legend(self):
        color_map = [
            (ProcessStep.StepState.FAILED, WBColor.GREEN_LIGHT.value),
            (ProcessStep.StepState.CANCELED, WBColor.BLUE_LIGHT.value),
        ]
        assert set((y.icon, y.label, y.value) for y in get_state_legend(color_map)[0].items) == {
            (
                WBColor.GREEN_LIGHT.value,
                ProcessStep.StepState.FAILED.label,
                ProcessStep.StepState.FAILED.value,
            ),
            (
                WBColor.BLUE_LIGHT.value,
                ProcessStep.StepState.CANCELED.label,
                ProcessStep.StepState.CANCELED.value,
            ),
        }

    def test_get_state_formatting(self):
        color_map = [
            (ProcessStep.StepState.WAITING, WBColor.GREY.value),
            (ProcessStep.StepState.ACTIVE, WBColor.YELLOW_DARK.value),
        ]
        assert set(
            (y.style["backgroundColor"], y.condition[1]) for y in get_state_formatting(color_map)[0].formatting_rules
        ) == {
            (WBColor.GREY.value, ProcessStep.StepState.WAITING.value),
            (WBColor.YELLOW_DARK.value, ProcessStep.StepState.ACTIVE.value),
        }



from datetime import date

import pytest
from faker import Faker
from pandas.tseries.offsets import BDay
from rest_framework.test import APIRequestFactory
from wbcore.configs.registry import ConfigRegistry
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.geography.tests.conftest import *

fake = Faker()


def app_pre_migration(sender, app_config, **kwargs):
    cur = connection.cursor()
    cur.execute("CREATE EXTENSION IF NOT EXISTS pg_trgm;")
    cur.execute("CREATE EXTENSION IF NOT EXISTS btree_gin;")
    cur.execute("CREATE EXTENSION IF NOT EXISTS btree_gist;")


@pytest.fixture(autouse=True, scope="session")
def django_test_environment(django_test_environment):
    from django.apps import apps

    get_models = apps.get_models

    for m in [m for m in get_models() if not m._meta.managed]:
        m._meta.managed = True


@pytest.fixture()
def super_user():
    return UserFactory.create(is_superuser=True)


@pytest.fixture()
def weekday() -> date:
    return (fake.date_object() - BDay(0)).date()


pre_migrate.connect(app_pre_migration)


@pytest.fixture()
def config_registry():
    request = APIRequestFactory().get("/")
    registry = ConfigRegistry(request)
    return registry


@pytest.fixture
def chrome_options(chrome_options):
    # chrome_options.add_argument("--headless=new")
    chrome_options.add_argument("--disable-extensions")
    return chrome_options



from wbcore.serializers import JSONTextEditorField
from wbcore.serializers.fields.types import ReturnContentType, WBCoreType


class TestJSONTextEditorField:
    def setup_method(self):
        self.field = JSONTextEditorField()

    def test_not_none(self):
        assert self.field is not None

    def test_field_type(self):
        assert self.field.field_type == WBCoreType.TEXTEDITOR.value

    def test_representation(self):
        assert self.field.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "content_type": ReturnContentType.JSON.value,
                "label": None,
                "type": self.field.field_type,
                "required": True,
                "read_only": False,
                "decorators": [],
                "depends_on": [],
            },
        )



# Generated by Django 4.1.10 on 2023-09-12 09:07

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("workflow", "0002_alter_step_step_type"),
    ]

    operations = [
        migrations.AlterField(
            model_name="condition",
            name="attribute_name",
            field=models.CharField(
                help_text="The name of the attribute on the attached instance. Its value will be evaluated against the condition.",
                max_length=64,
                verbose_name="Attribute Name",
            ),
        ),
        migrations.AlterField(
            model_name="condition",
            name="expected_value",
            field=models.CharField(
                help_text="The expected value of the attribute that will satisfy the condition.",
                max_length=64,
                verbose_name="Expected Value",
            ),
        ),
        migrations.AlterField(
            model_name="process",
            name="content_type",
            field=models.ForeignKey(
                blank=True,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="contenttypes.contenttype",
                verbose_name="Attached Model",
            ),
        ),
        migrations.AlterField(
            model_name="step",
            name="status",
            field=models.CharField(
                blank=True,
                help_text="The status that will be set in the attached instance's status field upon transitioning to this step. Only applicable if attached model is set.",
                max_length=64,
                null=True,
                verbose_name="Status",
            ),
        ),
        migrations.AlterField(
            model_name="workflow",
            name="single_instance_execution",
            field=models.BooleanField(
                default=False,
                help_text="Indicates wether the workflow is restricted to only one running workflow at a time. If an instance is attached to the workflow this will restrict the workflow execution per model instance.",
                verbose_name="Single Instance Execution",
            ),
        ),
        migrations.AlterField(
            model_name="workflow",
            name="status_field",
            field=models.CharField(
                blank=True,
                help_text="The status field name in the attached model that needs to be updated by the workflow",
                max_length=64,
                null=True,
                verbose_name="Status Field",
            ),
        ),
    ]



from django.db import models

from wbcore.contrib.dataloader.dataloaders.proxies import DataloaderProxy


class EntityQuerySet(models.QuerySet):
    """A custom QuerySet for entities that provides an interface to a dataloader proxy.

    This QuerySet subclass adds a method to interact with the associated dataloader
    proxy. It is designed to be used with models that inherit from the `Entity` class,
    providing a streamlined way to access dataloader functionality directly from
    QuerySet instances.
    """

    @property
    def dl(self) -> DataloaderProxy:
        """Provides access to the dataloader proxy for the entities in the QuerySet.

        This method allows for easy retrieval of the DataloaderProxy instance
        associated with the QuerySet. It enables the utilization of dataloader
        functionalities directly from the QuerySet, facilitating data fetching and
        processing tasks.

        Returns:
            DataloaderProxy: An instance of DataloaderProxy associated with the
                entities in the QuerySet.
        """
        return self.model.dl_proxy(self)



# Generated by Django 4.1.8 on 2023-04-13 10:33

import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("directory", "0001_initial"),
        ("authentication", "0001_initial_squashed"),
    ]

    operations = [
        migrations.AddField(
            model_name="user",
            name="profile",
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="user_account",
                to="directory.person",
            ),
        )
    ]



from contextlib import suppress
from dataclasses import dataclass
from typing import TYPE_CHECKING

from django.shortcuts import get_object_or_404
from rest_framework.fields import get_attribute

from wbcore.utils.views import parse_query_parameters_list

if TYPE_CHECKING:
    from django.db.models import Model


@dataclass
class CurrentUserDefault:
    """
    Returns the profile of the current user when called.
    """

    user_attr: str = None
    requires_context: str = True
    as_list: bool = False

    def __call__(self, serializer_instance):
        attr = None
        if request := serializer_instance.context.get("request"):
            attr = request.user
            if self.user_attr:
                try:
                    attr = get_attribute(attr, self.user_attr.split("."))
                except AttributeError:
                    attr = None
        if self.as_list and attr:
            attr = [attr]
        return attr


class DefaultFromGET:
    """
    Extracts the informations from the get request depending of the query_name.
    """

    def __init__(self, query_name: str, many: bool = False):
        self.many = many
        self.query_name = query_name

    requires_context = True

    def __call__(self, serializer_instance):
        if request := serializer_instance.context.get("request", None):
            data = request.GET.get(self.query_name, "")
            parameters_list = parse_query_parameters_list(data)

            if (not parameters_list or not self.many) and data:
                return data
            elif parameters_list:
                return parameters_list
        return None


class DefaultFromKwargs:
    """
    Extracts the informations from the view kwargs depending of the query_name.
    """

    def __init__(self, query_name: str):
        self.query_name = query_name

    requires_context = True

    def __call__(self, serializer_instance):
        if (view := serializer_instance.view) and (item_id := view.kwargs.get(self.query_name, None)):
            return item_id
        return None


class DefaultFromView:
    """
    Extracts the informations from the view kwargs depending of the query_name.
    """

    def __init__(self, source: str):
        self.source = source

    requires_context = True

    def __call__(self, serializer_instance):
        if view := serializer_instance.view:
            with suppress(AttributeError):
                return get_attribute(view, self.source.split("."))
        return None


class DefaultFromGetOrKwargs:
    """
    Extracts the informations from the get request or the view kwargs depending of the query_name.
    """

    def __init__(self, query_name_get: str, query_name_kwargs: str, many: bool = False):
        self.defaults = [DefaultFromGET(query_name_get, many), DefaultFromKwargs(query_name_kwargs)]

    requires_context = True

    def __call__(self, serializer_instance):
        for default in self.defaults:
            if result := default(serializer_instance):
                return result

        return None


class DefaultAttributeFromRemoteField(DefaultFromKwargs):
    def __init__(self, query_name: str, model: type["Model"], source: str | None = None):
        self.source = source
        self.model = model
        super().__init__(query_name)

    def __call__(self, serializer_instance):
        if item_id := super().__call__(serializer_instance):
            instance = get_object_or_404(self.model, pk=item_id)
            if self.source:
                try:
                    instance = get_attribute(instance, self.source.split("."))
                except AttributeError:
                    instance = None
            return instance


class DefaultAttributeFromObject:
    requires_context = True

    def __init__(self, source: str | None = None):
        self.source = source

    def __call__(self, serializer_instance):
        if (view := serializer_instance.view) and "pk" in view.kwargs:
            instance = view.get_object()
            if self.source:
                try:
                    instance = get_attribute(instance, self.source.split("."))
                except AttributeError:
                    instance = None
            return instance



import random

import factory
from faker import Faker

from wbcore.contrib.directory.serializers import PersonModelSerializer
from wbcore.contrib.workflow.models import Display

fake = Faker()


def _generate_grid_areas() -> list[list[str]]:
    outer_list = []
    inner_list_length = random.randint(1, 5)
    for k in range(random.randint(1, 5)):
        inner_list = []
        for i in range(inner_list_length):
            inner_list.append(random.choice(PersonModelSerializer.Meta.fields))
        outer_list.append(inner_list)
    return outer_list


class DisplayFactory(factory.django.DjangoModelFactory):
    name = factory.Faker("text", max_nb_chars=128)
    grid_template_areas = factory.LazyAttribute(lambda o: _generate_grid_areas())

    class Meta:
        model = Display



from django.db import models
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from wbcore.contrib.workflow.models.data import Data
from wbcore.models import WBModel


class Condition(WBModel):
    """A condition that needs to be satisfied in order to make the transition to the next workflow step"""

    class Operator(models.TextChoices):
        LT = "<", _("Less Than")
        LTE = "<=", _("Less Than or Equal")
        EQ = "==", _("Equal")
        GTE = ">=", _("Greater Than or Equal")
        GT = ">", _("Greater Than")

    transition = models.ForeignKey(
        to="workflow.Transition",
        verbose_name=_("Transition"),
        on_delete=models.CASCADE,
        related_name="associated_conditions",
    )
    attribute_name = models.CharField(
        max_length=64,
        verbose_name=_("Attribute Name"),
        help_text=_(
            "The name of the attribute on the attached instance. Its value will be evaluated against the condition."
        ),
    )
    operator = models.CharField(max_length=64, choices=Operator.choices, verbose_name=_("Operator"))
    expected_value = models.CharField(
        max_length=64,
        verbose_name=_("Expected Value"),
        help_text=_("The expected value of the attribute that will satisfy the condition."),
    )
    negate_operator = models.BooleanField(default=False, verbose_name=_("Negate Operator"))

    @property
    def errors(self):
        if not hasattr(self, "_errors"):
            raise ValueError(_('Please call "satisfied" before accessing errors'))
        return self._errors

    def satisfied(self, instance) -> bool:
        """Checks if the condition is satisfied for the provided instance

        Args:
            instance: The instance that should be checked.

        Returns:
            bool: True if the condition satisfies the provided instance, False otherwise
        """

        self._errors = []
        try:
            attribute_value = getattr(instance, self.attribute_name)
            try:
                # We need to cast the expected value to the type of the attached instance's field.
                # We technically do not need to do this for attached data as the data value is already a string, same as the expected value.
                casted_expected_value = Data.cast_value_from_target_object(attribute_value, self.expected_value)

                evaluated_condition = False
                if self.operator == self.Operator.EQ:
                    evaluated_condition = attribute_value == casted_expected_value
                elif self.operator == self.Operator.GT:
                    evaluated_condition = attribute_value > casted_expected_value
                elif self.operator == self.Operator.GTE:
                    evaluated_condition = attribute_value >= casted_expected_value
                elif self.operator == self.Operator.LT:
                    evaluated_condition = attribute_value < casted_expected_value
                elif self.operator == self.Operator.LTE:
                    evaluated_condition = attribute_value <= casted_expected_value

                if self.negate_operator:
                    return not evaluated_condition
                return evaluated_condition

            except ValueError:
                self._errors.append(
                    gettext(
                        "Condition error: Type of attribute ({}) and expected value ({}) uncompatible or wrong datetime format!"
                    ).format(type(attribute_value), self.expected_value)
                )
                return False
        except AttributeError:
            self._errors.append(
                gettext("Condition error: No attribute called {} found in attached instance or data!").format(
                    self.attribute_name
                )
            )
        return False

    def __str__(self) -> str:
        if self.negate_operator:
            return _("{} not {} {}").format(self.attribute_name, self.operator, self.expected_value)
        return f"{self.attribute_name} {self.operator} {self.expected_value}"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:workflow:condition"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:workflow:conditionrepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{transition}}: {{attribute_name}} {{negate_operator}} {{operator}} {{expected_value}}"

    class Meta:
        verbose_name = _("Condition")
        verbose_name_plural = _("Conditions")
        constraints = [
            models.UniqueConstraint(
                fields=["transition", "attribute_name", "operator", "negate_operator", "expected_value"],
                name="unique_transition_attribute_name_operator_negate_expected_value",
            ),
        ]



from django.db import models
from django.utils.translation import gettext_lazy as _
from mptt.models import MPTTModel, TreeForeignKey

from wbcore.models import WBModel
from wbcore.utils.models import ComplexToStringMixin


class DocumentType(ComplexToStringMixin, WBModel, MPTTModel):
    name = models.CharField(max_length=255)
    parent = TreeForeignKey(
        "self", on_delete=models.CASCADE, null=True, blank=True, related_name="children", verbose_name=_("Parent")
    )
    system_key = models.CharField(max_length=255, null=True, blank=True)

    def compute_str(self):
        if self.id:
            ancestors_names = self.get_ancestors(ascending=False, include_self=True).values_list("name", flat=True)
            return " -> ".join(list(ancestors_names))
        return self.name

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:documents:documenttype"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:documents:documenttyperepresentation-list"

    class Meta:
        verbose_name = _("Document Type")
        verbose_name_plural = _("Document Types")



import pytest
from django.test import Client, TestCase
from django.urls import reverse
from wbcore.contrib.authentication.factories import SuperUserFactory
from wbcore.contrib.example_app.factories import TeamResultsFactory
from wbcore.contrib.example_app.serializers import TeamResultsModelSerializer
from wbcore.contrib.example_app.tests.test_viewsets.test_utils_viewsets import (
    get_create_view,
    get_delete_view,
    get_detail_view,
    get_partial_view,
    get_update_view,
)
from wbcore.contrib.example_app.viewsets import TeamResultsModelViewSet


@pytest.mark.django_db
class TestTeamResultsModelViewSet(TestCase):
    def setUp(self):
        self.user = SuperUserFactory()
        self.client = Client()
        self.client.force_login(user=self.user)
        self.list_url = reverse("example_app:teamresults-list")
        self.detail_url_str = "example_app:teamresults-detail"

    def test_list_view(self):
        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)

    def test_create_view(self):
        team_result = TeamResultsFactory()
        response = get_create_view(self.client, team_result, self.user, self.list_url, TeamResultsModelViewSet)
        # It is not possible to create an team results, since the get_endpoint_url returns None.
        self.assertEqual(response.status_code, 405)

    def test_detail_view(self):
        instance = TeamResultsFactory()
        response = get_detail_view(self.client, instance.pk, self.detail_url_str)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["id"], instance.id)

    def test_update_view(self):
        instance = TeamResultsFactory()
        instance.points = 5
        response = get_update_view(self.client, instance, TeamResultsModelSerializer, self.detail_url_str)
        # It is not possible to update an team results, since the get_endpoint_url returns None.
        self.assertEqual(response.status_code, 405)

    def test_partial_update_view(self):
        instance = TeamResultsFactory()
        response = get_partial_view(self.client, instance.id, {"points": 5}, self.detail_url_str)
        self.assertEqual(response.status_code, 405)

    def test_delete_view(self):
        instance = TeamResultsFactory()
        response = get_delete_view(self.client, self.detail_url_str, instance.pk)
        # It is not possible to delete an team results, since the get_endpoint_url returns None.
        self.assertEqual(response.status_code, 405)



import pytest
from rest_framework.reverse import reverse

from wbcore.contrib.notifications.serializers import (
    NotificationTypeRepresentationSerializer,
    NotificationTypeSettingModelSerializer,
)


@pytest.mark.django_db
class TestNotificationTypeRepresentationSerializer:
    def test_serialize_instance(self, notification_type):
        serializer = NotificationTypeRepresentationSerializer(notification_type)
        assert serializer.data == {
            "id": notification_type.id,
            "code": notification_type.code,
            "title": notification_type.title,
            "help_text": notification_type.help_text,
        }


@pytest.mark.django_db
class TestNotificationTypeSettingModelSerializer:
    def test_serialize_instance(self, notification_type_setting, request_with_user):
        # We need to add the help_text here manually - we assume the queryset passed into the serializer annotates it
        # in the viewset.
        notification_type_setting.help_text = notification_type_setting.notification_type.help_text
        serializer = NotificationTypeSettingModelSerializer(
            notification_type_setting, context={"request": request_with_user}
        )

        assert serializer.data == {
            "id": notification_type_setting.id,
            "notification_type": notification_type_setting.notification_type.id,
            "_notification_type": NotificationTypeRepresentationSerializer(
                notification_type_setting.notification_type
            ).data,
            "help_text": notification_type_setting.notification_type.help_text,
            "user": notification_type_setting.user.id,
            "enable_web": notification_type_setting.enable_web,
            "enable_mobile": notification_type_setting.enable_mobile,
            "enable_email": notification_type_setting.enable_email,
            "_additional_resources": {},
            "_update_url": reverse(
                "wbcore:notifications:notification_type_setting-list", args=[], request=request_with_user
            ),
            "locked_icon": None,
        }

    def test_deserialize_partial_instance(self, notification_type_setting, request_with_user):
        serializer = NotificationTypeSettingModelSerializer(
            notification_type_setting,
            data={"enable_web": True},
            partial=True,
            context={"request": request_with_user},
        )

        assert serializer.is_valid()
        assert serializer.validated_data["enable_web"]  # type: ignore



import pytest
from django.utils.translation import gettext
from pytest_mock import MockerFixture
from wbcore.contrib.workflow.models import FinishStep, Process, ProcessStep, Step


class TestFinishStep:
    @pytest.fixture(autouse=True)
    def patch_super_finish(self, mocker: MockerFixture):
        mocker.patch("wbcore.contrib.workflow.models.step.Step.finish", autospec=True)

    @pytest.fixture
    def mocked_finish_step(self, mocker: MockerFixture):
        return mocker.MagicMock(spec=FinishStep)

    @pytest.fixture
    def mocked_instance(self, mocker: MockerFixture):
        return mocker.MagicMock(first_name="Original")

    @pytest.fixture
    def mocked_process(self, mocker: MockerFixture, mocked_instance):
        mocked_process = mocker.MagicMock(
            finished=None,
            instance=mocked_instance,
            save=mocker.MagicMock(),
            spec=Process,
            state=Process.ProcessState.ACTIVE,
        )
        return mocked_process

    @pytest.fixture
    def mocked_process_step(self, mocker: MockerFixture, mocked_process):
        return mocker.MagicMock(spec=ProcessStep, process=mocked_process)

    def test_run(self, mocker: MockerFixture, mocked_process_step, mocked_finish_step):
        # Arrange
        mocked_qs = mocker.patch("wbcore.contrib.workflow.models.step.ProcessStep.objects.filter")
        mocked_qs.return_value.exclude.return_value.exists.return_value = False
        mocked_set_finished = mocker.patch.object(mocked_finish_step, "set_finished")
        # Act
        FinishStep.run(mocked_finish_step, mocked_process_step)
        # Assert
        mocked_set_finished.assert_called_once_with(mocked_process_step)

    def test_run_failed(self, mocker: MockerFixture, mocked_process_step, mocked_finish_step):
        # Arrange
        mocked_qs = mocker.patch("wbcore.contrib.workflow.models.step.ProcessStep.objects.filter")
        mocked_qs.return_value.exclude.return_value.exists.return_value = True
        mocked_set_failed = mocker.patch.object(mocked_finish_step, "set_failed")
        # Act
        FinishStep.run(mocked_finish_step, mocked_process_step)
        # Assert
        mocked_set_failed.assert_called_once_with(
            mocked_process_step,
            gettext("There are process steps still running for this workflow!"),
        )

    @pytest.mark.parametrize(
        "unfinished_state",
        [
            ProcessStep.StepState.ACTIVE,
            ProcessStep.StepState.WAITING,
            ProcessStep.StepState.CANCELED,
            ProcessStep.StepState.FAILED,
        ],
    )
    def test_finish_not_finished(
        self,
        mocker: MockerFixture,
        mocked_process,
        mocked_process_step,
        unfinished_state,
    ):
        # Arrange
        mocked_process.preserved_instance = {"first_name": "Test"}
        mocked_process_step.state = unfinished_state

        finish_step = FinishStep()
        mock_serializer = mocker.patch("wbcore.contrib.workflow.models.step.get_model_serializer_class_for_instance")
        # Act
        finish_step.finish(mocked_process_step)
        # Assert
        mock_serializer.assert_not_called()
        mocked_process_step.process.save.assert_not_called()
        assert mocked_process.finished is None
        assert mocked_process.state == Process.ProcessState.ACTIVE

    def test_finish_write_preserved_instance(
        self,
        mocker: MockerFixture,
        mocked_process,
        mocked_process_step,
    ):
        # Arrange
        preserved_data = {"first_name": "Updated"}

        mocked_serializer = mocker.MagicMock()
        mocked_serializer.validated_data = preserved_data
        mocked_serializer_class = mocker.MagicMock()
        mocked_serializer_class.return_value = mocked_serializer
        mocker.patch(
            "wbcore.contrib.workflow.models.step.get_model_serializer_class_for_instance",
            return_value=mocked_serializer_class,
        )
        mocked_serializer.is_valid.return_value = True
        mocked_serializer.update = mocker.MagicMock()

        mocked_process.preserved_instance = preserved_data

        mocked_step = mocker.MagicMock(spec=Step, write_preserved_instance=True)
        mocked_step.workflow.preserve_instance = True

        mocked_process_step.state = ProcessStep.StepState.FINISHED
        mocked_process_step.step = mocked_step

        finish_step = FinishStep()
        mocker.patch.object(finish_step, "write_preserved_instance", True)
        # Act
        finish_step.finish(mocked_process_step)
        # Assert
        mocked_serializer.is_valid.assert_called_once()
        mocked_serializer.update.assert_called_once_with(mocked_process.instance, preserved_data)
        mocked_process.save.assert_called_once()
        assert mocked_process.finished is not None
        assert mocked_process.state == Process.ProcessState.FINISHED

    def test_finish_do_not_write_mocker(self, mocker: MockerFixture, mocked_process, mocked_process_step):
        # Arrange
        mocked_process.preserved_instance = {"first_name": "Updated"}
        mocked_process.state = Process.ProcessState.ACTIVE

        mocked_process_step.state = ProcessStep.StepState.FINISHED

        finish_step = FinishStep()
        mocker.patch.object(finish_step, "write_preserved_instance", False)

        mock_serializer = mocker.patch("wbcore.contrib.workflow.models.step.get_model_serializer_class_for_instance")

        # Act
        finish_step.finish(mocked_process_step)

        # Assert
        mock_serializer.assert_not_called()
        mocked_process.save.assert_called_once()
        assert mocked_process.finished is not None
        assert mocked_process.state == Process.ProcessState.FINISHED

    def test_finish_no_preserved_instance_mocker(self, mocker: MockerFixture, mocked_process, mocked_process_step):
        # Arrange
        mocked_process.preserved_instance = None
        mocked_process.state = Process.ProcessState.ACTIVE

        mocked_process_step.state = ProcessStep.StepState.FINISHED

        finish_step = FinishStep()
        mocker.patch.object(finish_step, "write_preserved_instance", True)

        mock_serializer = mocker.patch("wbcore.contrib.workflow.models.step.get_model_serializer_class_for_instance")

        # Act
        finish_step.finish(mocked_process_step)

        # Assert
        mock_serializer.assert_not_called()
        mocked_process.save.assert_called_once()
        assert mocked_process.finished is not None
        assert mocked_process.state == Process.ProcessState.FINISHED



from typing import Optional

from django.utils.translation import gettext as _

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class RevisionDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="date_created", label=_("Created At")),
                dp.Field(key="user", label=_("User")),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [["date_created", "user"], [repeat_field(2, "versions_section")]],
            [create_simple_section("versions_section", _("Versions"), [["versions"]], "versions", collapsed=False)],
        )


class VersionDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="id", label=_("ID")),
                dp.Field(key="revision", label=_("Revision")),
                dp.Field(key="date_created", label=_("Creation Date")),
                # dp.Field(key="object_repr", label=_("Object repr")),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [["revision", "date_created"], ["object_repr", "object_id"], [repeat_field(2, "serialized_data")]]
        )



import pytest
from pytest_factoryboy import register
from wbcore.contrib.authentication.factories import SuperUserFactory, UserFactory
from wbcore.contrib.directory.factories import PersonFactory
from wbcore.tests.conftest import *  # noqa

from ..factories import (
    CrontabScheduleFactory,
    DataBackendFactory,
    ExportSourceFactory,
    ImportCredentialFactory,
    ImportModelFactory,
    ImportSourceFactory,
    ParserHandlerFactory,
    ProviderFactory,
    SourceFactory,
)

register(CrontabScheduleFactory)
register(ImportSourceFactory)
register(ImportCredentialFactory)
register(SourceFactory)
register(ParserHandlerFactory)
register(ExportSourceFactory)
register(ImportModelFactory)
register(ProviderFactory)
register(DataBackendFactory)
register(UserFactory)
register(SuperUserFactory, "superuser")
register(PersonFactory)


@pytest.fixture(
    autouse=True, scope="session"
)  # Might want to find a way to registered default conftest logic automatically
def django_test_environment(django_test_environment):
    from django.apps import apps

    get_models = apps.get_models

    for m in [m for m in get_models() if not m._meta.managed]:
        m._meta.managed = True



from itertools import chain

from django.contrib.contenttypes.models import ContentType
from django.db.models import Q

from wbcore import filters
from wbcore.content_type.utils import (
    get_ancestors_content_type,
    get_view_content_type_id,
)

from .models import Tag


class TagFilterMixin(filters.FilterSet):
    tags = filters.ModelMultipleChoiceFilter(
        endpoint=Tag.get_representation_endpoint(),
        value_key=Tag.get_representation_value_key(),
        label_key=Tag.get_representation_label_key(),
        label="Tags",
        filter_params=lambda request, view: {"content_type": get_view_content_type_id(view)},
        queryset=Tag.objects.all(),
    )


class TagFilterSet(filters.FilterSet):
    content_type = filters.ModelMultipleChoiceFilter(
        endpoint="wbcore:contenttyperepresentation-list",
        value_key="id",
        label_key="{{app_label}} | {{model}}",
        queryset=ContentType.objects.all(),
        method="filter_content_type",
    )

    def filter_content_type(self, queryset, label, value):
        if value:
            content_types = list(chain(*[get_ancestors_content_type(content_type) for content_type in value]))
            return queryset.filter(Q(content_type__isnull=True) | Q(content_type__in=content_types))
        return queryset

    class Meta:
        model = Tag
        fields = {"title": ["icontains", "exact"], "description": ["icontains"], "groups": ["exact"]}



from wbcore.contrib.guardian.filters import ObjectPermissionsFilter
from wbcore.viewsets.mixins import FilterMixin


class GuardianFilterMixin(FilterMixin):
    filter_backends = (ObjectPermissionsFilter, *FilterMixin.filter_backends)



from .apps import Apps, DevApps
from .authentication import Authentication
from .base import Base
from .celery import Celery, CeleryTest
from .i18nl10n import I18NL10N
from .mail import ConsoleEmail
from .media import LocalMedia, S3Media
from .middleware import DevMiddleware, Middleware
from .network import Network, SSLNetwork
from .rest_framework import Restframework
from .static import LocalStaticfiles, S3Staticfiles
from .templates import Templates
from .uvicorn import Uvicorn
from .wbcore import WBCore
from .maintenance import Maintenance
from .cache import Cache



import ast
from urllib.parse import unquote

from django import forms
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.forms import modelformset_factory
from django.forms.models import BaseModelFormSet
from psycopg.types.range import DateRange, TimestamptzRange


class NonrelatedInlineFormSet(BaseModelFormSet):
    """
    A basic implementation of an inline formset that doesn't make assumptions
    about any relationship between the form model and its parent instance.
    """

    def __init__(self, instance=None, save_as_new=None, **kwargs):
        self.instance = instance
        super().__init__(**kwargs)
        self.queryset = self.real_queryset

    @classmethod
    def get_default_prefix(cls):
        opts = cls.model._meta
        return opts.app_label + "-" + opts.model_name

    def save_new(self, form, commit=True):
        obj = super().save_new(form, commit=False)
        self.save_new_instance(self.instance, obj)
        if commit:
            obj.save()
        return obj


def nonrelated_inlineformset_factory(
    model, obj=None, queryset=None, formset=NonrelatedInlineFormSet, save_new_instance=None, **kwargs
):
    """
    FormSet factory that sets an explicit queryset on new classes.
    """
    FormSet = modelformset_factory(model, formset=formset, **kwargs)
    FormSet.real_queryset = queryset
    FormSet.save_new_instance = save_new_instance
    return FormSet


class ContentTypeMultiValueField(forms.fields.MultiValueField):
    def __init__(self, fields=None, *args, **kwargs):
        if fields is None:
            fields = (forms.IntegerField(), forms.IntegerField())
        super().__init__(fields, *args, **kwargs)

    def clean(self, value):
        value = self.to_python(value)
        res = []
        for val in value:
            res.append(super().clean(val))
        return res

    def to_python(self, value):
        if value:
            return ast.literal_eval(value)
        return []

    def compress(self, data_list):
        content_type_id, object_id = data_list
        try:
            content_type = ContentType.objects.get_for_id(content_type_id)
            return content_type.get_object_for_this_type(id=object_id)
        except (ContentType.DoesNotExist, content_type.model_class().DoesNotExist):
            return None


class TupleWidget(forms.TextInput):
    """
    Format a input of the form "param=val1,val2"
    """

    def value_from_datadict(self, data, files, name):
        if res := data.get(name, None):
            return unquote(res).split(",")


class DateRangeField(forms.MultiValueField):
    """
    A date range
    """

    default_error_messages = {
        "invalid": "Enter a valid date range.",
    }
    widget = TupleWidget
    form_date_field = forms.DateField
    python_field = DateRange

    def validate(self, value):
        if value.lower and value.upper and value.upper < value.lower:
            raise ValidationError(
                self.error_messages["invalid"],
                code="invalid_choice",
            )

    def __init__(self, **kwargs):
        fields = (
            self.form_date_field(required=True),
            self.form_date_field(required=True),
        )
        super().__init__(fields, **kwargs)

    def compress(self, values):
        try:
            lower, upper = values
            return self.python_field(lower=lower, upper=upper)
        except ValueError:
            return None


class DateTimeRangeField(DateRangeField):
    """
    A date time range
    """

    form_date_field = forms.DateTimeField
    python_field = TimestamptzRange



from datetime import date, datetime, timedelta

import pytest
from django.forms.models import model_to_dict
from rest_framework.exceptions import ValidationError
from rest_framework.test import APITestCase
from wbcore.contrib.example_app.factories import LeagueFactory, SportFactory
from wbcore.contrib.example_app.serializers import LeagueModelSerializer


@pytest.mark.django_db
class TestLeagueModelSerializer(APITestCase):
    def test_league_serializer(self):
        sport = SportFactory()
        league_data: dict = model_to_dict(LeagueFactory.build(sport=sport))
        league_serializer = LeagueModelSerializer(data=league_data)

        self.assertTrue(league_serializer.is_valid())

    def test_league_exists(self):
        sport = SportFactory()
        league = LeagueFactory(sport=sport)
        new_league_data: dict = model_to_dict(LeagueFactory.build(sport=sport, name=league.name))

        with pytest.raises(ValidationError):
            LeagueModelSerializer(data=new_league_data).is_valid(raise_exception=True)

    def test_league_wrong_date(self):
        sport = SportFactory()
        future_date: date = (datetime.now() + timedelta(days=1)).date()
        league_data: dict = model_to_dict(LeagueFactory.build(sport=sport, established_date=future_date))

        with pytest.raises(ValidationError):
            LeagueModelSerializer(data=league_data).is_valid(raise_exception=True)



import json
import operator
from functools import reduce

import django_filters
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django_filters.constants import EMPTY_VALUES

from wbcore.filters.mixins import WBCoreFilterMixin
from wbcore.forms import ContentTypeMultiValueField


class MultipleChoiceContentTypeFilter(WBCoreFilterMixin, django_filters.Filter):
    field_class = ContentTypeMultiValueField
    filter_type = "text"

    def _validate_initial_with_request(self, initial, request, name):
        initial = super()._validate_initial_with_request(initial, request, name)
        if initial:
            initial = json.dumps(initial)
        return initial

    def __init__(self, object_id_label="object_id", content_type_label="content_type", **kwargs):
        self.object_id_label = object_id_label
        self.content_type_label = content_type_label
        super().__init__(**kwargs)

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        conditions = [
            (
                Q(**{self.content_type_label: ContentType.objects.get_for_model(val)})
                & Q(**{self.object_id_label: val.id})
            )
            for val in value
        ]
        qs = qs.filter(reduce(operator.or_, conditions))

        if self.distinct:
            qs = qs.distinct()
        return qs



import pytest
from django.conf import settings

from wbcore.configs.decorators import register_config
from wbcore.configs.registry import ConfigRegistry
from wbcore.configs.views import ConfigAPIView


def test_registry(config_registry: ConfigRegistry):
    configs = config_registry.get_config_dict()
    assert configs.keys()


def test_release_note_config(config_registry: ConfigRegistry):
    release_notes = config_registry.get_config_dict()["release_notes"]
    assert release_notes["endpoint"]
    assert release_notes["unread_release_notes"]


def test_menu_config(config_registry: ConfigRegistry):
    menu = config_registry.get_config_dict()["menu"]
    assert menu


def test_share_config(config_registry: ConfigRegistry):
    share = config_registry.get_config_dict()["share"]
    assert share


def test_markdown_config(config_registry: ConfigRegistry):
    markdown = config_registry.get_config_dict()["markdown"]
    assert markdown["blockdiag"]


def test_menu_calendar_config(config_registry: ConfigRegistry):
    menu_calendar = config_registry.get_config_dict()["menu_calendar"]
    assert menu_calendar


@pytest.mark.parametrize("text, version", [("Foo bar", "Foo")])
def test_beta_button_config(config_registry: ConfigRegistry, text, version):
    settings.BETA_BUTTON_VERSION = version
    settings.BETA_BUTTON_TEXT = text
    beta_calendar = config_registry.get_config_dict()["beta_button"]
    assert beta_calendar["url"] == f"{settings.CDN_BASE_ENDPOINT_URL}/{version}/main.js"
    assert beta_calendar["text"] == text


def test_config_registry_decorator():
    def some_callable():
        pass

    assert not hasattr(some_callable, "_is_config")
    some_callable = register_config(some_callable)
    assert getattr(some_callable, "_is_config")


def test_config_view(rf):
    request = rf.get("/")
    view = ConfigAPIView.as_view()
    response = view(request)

    assert response.status_code == 200



from rest_framework.filters import BaseFilterBackend


class ObjectPermissionsFilter(BaseFilterBackend):
    """
    A filter backend that limits results to those where the requesting user
    has read object level permissions.
    """

    def filter_queryset(self, request, queryset, view):
        from guardian.shortcuts import get_objects_for_user
        from wbcore.contrib.guardian.models.mixins import PermissionObjectModelMixin

        model_class = queryset.model
        if issubclass(model_class, PermissionObjectModelMixin):
            user = request.user

            return get_objects_for_user(
                user, [model_class.view_perm_str], queryset, **model_class.guardian_shortcut_kwargs
            )
        return queryset



from django.utils.translation import gettext as _

from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

PROCESS_MENUITEM = MenuItem(
    label=_("Process"),
    endpoint="wbcore:workflow:process-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["workflow.view_process"]
    ),
)
PROCESSSTEP_MENUITEM = MenuItem(
    label=_("Process Step"),
    endpoint="wbcore:workflow:processstep-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["workflow.view_processstep"]
    ),
)



from .document_model_relationships import DocumentModelRelationship
from .document_types import DocumentType
from .documents import Document
from .shareable_links import ShareableLink, ShareableLinkAccess



from .contacts import PENDING_BANKINGCONTACT_MENUITEM, TELEPHONECONTACTSEARCH_MENUITEM
from .entries import (
    BANK_MENUITEM,
    COMPANY_MENUITEM,
    PERSON_MENUITEM,
    SYSTEMEMPLOYEE_MENUITEM,
    USERISMANAGER_MENUITEM,
    YOUR_CONTACT_MENUITEM,
)
from .relationships import CLIENTMANAGER_MENUITEM, RELATIONSHIPTYPE_MENUITEM
from .utils import (
    COMPANYTYPE_MENUITEM,
    CUSTOMERSTATUS_MENUITEM,
    POSITION_MENUITEM,
    SPECIALIZATION_MENUITEM,
)



from enum import Enum


class ChoiceEnum(Enum):
    @classmethod
    def choices(cls):
        return [(choice.name, choice.value) for choice in cls]

    @classmethod
    def model_choices(cls):
        return [(choice.value, choice.value) for choice in cls]

    @classmethod
    def names(cls):
        return [choice.name for choice in cls]

    @classmethod
    def values(cls):
        return [choice.value for choice in cls]

    @classmethod
    def to_dict(cls):
        return {choice.name: choice.value for choice in cls}



from django.urls import path

from wbcore.contrib.gleap.views import GleapAPITokenAPIView, GleapUserIdentityAPIView

urlpatterns = [
    path("user_identity/", GleapUserIdentityAPIView.as_view(), name="user_identity"),
    path("api_token/", GleapAPITokenAPIView.as_view(), name="api_token"),
]



from wbcore.configs.registry import ConfigRegistry


def test_authentication_config(config_registry: ConfigRegistry):
    authentication = config_registry.get_config_dict()["authentication"]
    assert authentication



import warnings

warnings.warn(
    "The 'wbcore.pandas' module is deprecated and will be removed in a future version. "
    "Please use 'wbcore.contrib.pandas' instead.",
    DeprecationWarning,
    stacklevel=2,
)
from wbcore.contrib.pandas.fields import (  # noqa
    PKField,
    CharField,
    DateField,
    DateRangeField,
    BooleanField,
    TextField,
    EmojiRatingField,
    FloatField,
    IntegerField,
    YearField,
    ListField,
    JsonField,
    SparklineField,
    PandasFields,
)



# Generated by Django 5.0.3 on 2024-04-30 08:18

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("io", "0004_alter_importsource_status_exportsource"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name="exportsource",
            name="data",
            field=models.JSONField(blank=True, default=dict, null=True),
        ),
        migrations.AlterField(
            model_name="exportsource",
            name="query_str",
            field=models.TextField(blank=True, verbose_name="SQL query to be executed"),
        ),
        migrations.AlterField(
            model_name="exportsource",
            name="resource_path",
            field=models.CharField(
                blank=True, default="", max_length=255, null=True, verbose_name="Resource path to use when exporting"
            ),
        ),
        migrations.AlterField(
            model_name="importsource",
            name="data",
            field=models.JSONField(blank=True, default=dict, null=True),
        ),
        migrations.AddConstraint(
            model_name="exportsource",
            constraint=models.CheckConstraint(
                check=models.Q(
                    models.Q(("query_str__isnull", False), ("resource_path__isnull", False)),
                    models.Q(("data__exact", {}), _negated=True),
                    _connector="OR",
                ),
                name="check_either_data_or_resource_isnotnull",
            ),
        ),
        migrations.AlterField(
            model_name="exportsource",
            name="data",
            field=models.JSONField(blank=True, default=dict, null=True),
        ),
        migrations.AlterField(
            model_name="exportsource",
            name="format",
            field=models.IntegerField(
                choices=[(0, "CSV"), (1, "XLSX"), (2, "TSV"), (3, "JSON"), (4, "YAML")],
                default=0,
                verbose_name="Format of file to be exported",
            ),
        ),
        migrations.AlterField(
            model_name="importsource",
            name="data",
            field=models.JSONField(blank=True, default=dict, null=True),
        ),
    ]



import re

from django.core.exceptions import ValidationError
from django.db.models import CharField


def validate_color(color_string):
    """Hexadecimal validation."""
    if color_string is None:
        return color_string

    if not re.search(r"^#(?:[0-9a-fA-F]{3}){1,2}$", color_string):
        raise ValidationError("Please provide a valid hexadecimal color.")

    return color_string


class ColorField(CharField):
    description = "A field storing a hexadecimal color value"
    default_validators = [validate_color]

    def __init__(self, *args, **kwargs):
        kwargs["max_length"] = 7
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["max_length"]
        return name, path, args, kwargs



from django_better_admin_arrayfield.forms.widgets import DynamicArrayWidget


class DynamicArrayColorWidget(DynamicArrayWidget):
    template_name = "wbcore/dynamic_color_array.html"

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        context["default"] = "#ffffff"
        return context

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from datetime import timedelta

from configurations import values


class Authentication:
    AUTHENTICATION_BACKENDS = (
        "django.contrib.auth.backends.ModelBackend",  # this is default
        "guardian.backends.ObjectPermissionBackend",
    )

    AUTH_PASSWORD_VALIDATORS = [
        {"NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"},
        {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
        {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
        {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
    ]

    JWT_AUTH = {"JWT_AUTH_COOKIE": "JWT"}
    JWT_ACCESS_TOKEN_LIFETIME = values.FloatValue(5 * 60, environ_prefix=None)  # in seconds
    JWT_REFRESH_TOKEN_LIFETIME = values.FloatValue(24 * 60 * 60, environ_prefix=None)  # in seconds
    JWT_SLIDING_TOKEN_LIFETIME = values.FloatValue(5 * 60, environ_prefix=None)  # in seconds
    JWT_SLIDING_TOKEN_REFRESH_LIFETIME = values.FloatValue(24 * 60 * 60, environ_prefix=None)  # in seconds
    JWT_COOKIE_KEY = values.Value("JWT-access", environ_prefix=None)

    @property
    def SIMPLE_JWT(self):
        return {
            "ACCESS_TOKEN_LIFETIME": timedelta(seconds=self.JWT_ACCESS_TOKEN_LIFETIME),
            "REFRESH_TOKEN_LIFETIME": timedelta(seconds=self.JWT_REFRESH_TOKEN_LIFETIME),
            "ALGORITHM": "HS256",
            "SIGNING_KEY": self.SECRET_KEY,
            "ROTATE_REFRESH_TOKENS": False,
            "BLACKLIST_AFTER_ROTATION": True,
            "VERIFYING_KEY": None,
            "AUTH_HEADER_TYPES": ("Bearer",),
            "USER_ID_FIELD": "id",
            "USER_ID_CLAIM": "user_id",
            "AUTH_TOKEN_CLASSES": ("rest_framework_simplejwt.tokens.AccessToken",),
            "TOKEN_TYPE_CLAIM": "token_type",
            "SLIDING_TOKEN_REFRESH_EXP_CLAIM": "refresh_exp",
            "SLIDING_TOKEN_LIFETIME": timedelta(seconds=self.JWT_SLIDING_TOKEN_LIFETIME),
            "SLIDING_TOKEN_REFRESH_LIFETIME": timedelta(seconds=self.JWT_SLIDING_TOKEN_REFRESH_LIFETIME),
        }



from collections import defaultdict
from dataclasses import dataclass
from typing import List, Literal

import pandas as pd

from wbcore.serializers.fields.types import DisplayMode


@dataclass
class _BaseField:
    key: str
    label: str
    type: str = None
    decorators: List = None
    help_text: str = None

    def to_dict(self):
        base = {"key": self.key, "label": self.label, "type": self.type}
        if self.decorators:
            base["decorators"] = self.decorators

        for _attr in ["help_text", "extra"]:
            attr = getattr(self, _attr, None)
            if attr:
                base[_attr] = attr
        return base

    def to_representation(self, value: pd.Series) -> pd.Series:
        return value


@dataclass
class PKField(_BaseField):
    type: str = "primary_key"


@dataclass
class CharField(_BaseField):
    type: str = "text"

    def to_representation(self, value: pd.Series) -> pd.Series:
        return super().to_representation(value).astype("string", errors="ignore")


@dataclass
class DateField(_BaseField):
    type: str = "date"

    def to_representation(self, value: pd.Series) -> pd.Series:
        return pd.to_datetime(super().to_representation(value)).dt.strftime("%Y-%m-%d")


@dataclass
class DateRangeField(_BaseField):
    type: str = "daterange"


@dataclass
class BooleanField(_BaseField):
    type: str = "boolean"


@dataclass
class TextField(_BaseField):
    type: str = "texteditor"

    def to_representation(self, value: pd.Series) -> pd.Series:
        return super().to_representation(value).astype("string", errors="ignore")


@dataclass
class EmojiRatingField(_BaseField):
    type: str = "emojirating"


@dataclass
class FloatField(_BaseField):
    type: str = "number"
    precision: int = 2
    percent: bool = False
    display_mode: DisplayMode = None
    disable_formatting: bool = False

    def to_dict(self):
        base = super().to_dict()
        base.update(
            {
                "precision": self.precision,
            }
        )
        if self.percent:
            base["type"] = "percent"
        if self.display_mode:
            base["display_mode"] = self.display_mode.value
        base["disable_formatting"] = self.disable_formatting
        return base

    def to_representation(self, value: pd.Series) -> pd.Series:
        return super().to_representation(value).astype("float", errors="ignore")


@dataclass
class IntegerField(FloatField):
    type: str = "number"
    precision: int = 0

    def to_representation(self, value: pd.Series) -> pd.Series:
        return super().to_representation(value).astype("Int64", errors="ignore")


@dataclass
class YearField(IntegerField):
    precision: int = 0
    disable_formatting: bool = True

    def __post_init__(self):
        self.precision = 0


@dataclass
class ListField(_BaseField):
    type: str = "list"


@dataclass
class JsonField(_BaseField):
    type: str = "json"


@dataclass
class SparklineField(ListField):
    type: str = "sparkline"
    dimension: Literal["single"] | Literal["double"] = (
        "single"  # "single" for data [y1, y2, y3... ] or "double" if contain already the X axis as [[x1, y1], [x2, y2], ... ]
    )

    def to_dict(self):
        rv = super().to_dict()
        rv["sparkline_type"] = "bar"
        return rv

    def _sanitize_row(self, row):
        if not row:
            row = [[]]  # if row is [] or null, we default to an empty list of list
        if self.dimension == "single":  # This ensure that the returned data format contains a list of tuple of x,y
            row = list(map(lambda o: (o[0], o[1]), enumerate(row)))
        return row

    def to_representation(self, value: pd.Series) -> pd.Series:
        return super().to_representation(value).apply(lambda x: self._sanitize_row(x))


@dataclass(unsafe_hash=True)
class PandasFields:
    fields: List[_BaseField]

    def to_dict(self):
        fields = defaultdict(dict)

        for field in self.fields:
            fields[field.key] = field.to_dict()

        return fields



from django.utils.translation import gettext as _

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import Display, Layout, Page, Style
from wbcore.metadata.configs.display.instance_display.operators import default, lte
from wbcore.metadata.configs.display.list_display import ListDisplay
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

EVENT_TYPE_FIELDS = [
    dp.Field(key="name", label=_("Name")),
    dp.Field(key="sport", label=_("Sport")),
    dp.Field(key="points", label=_("Points")),
    dp.Field(key="icon", label=_("Icon")),
    dp.Field(key="color", label=_("Color")),
]
EVENT_FIELDS = [
    dp.Field(key="match", label=_("Match")),
    dp.Field(key="event_type", label=_("Event Type")),
    dp.Field(key="person", label=_("Person")),
    dp.Field(key="minute", label=_("Minute")),
]


class EventDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        return dp.ListDisplay(fields=EVENT_FIELDS)

    def get_instance_display(self) -> Display:
        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["match", "event_type"],
                                ["person", "minute"],
                                ["event_description", "event_description"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(650): Layout(
                            grid_template_areas=[
                                ["match"],
                                ["event_type"],
                                ["person"],
                                ["minute"],
                                ["event_description"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
            ]
        )


class LeaguePlayerStatisticsDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        return dp.ListDisplay(
            fields=(dp.Field(key="person_name", label=_("Person")), dp.Field(key="count", label=_(" "))),
        )


class LeagueTeamStatisticsDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        return dp.ListDisplay(
            fields=(dp.Field(key="team_name", label=_("Team")), dp.Field(key="count", label=_(" "))),
        )


class EventMatchDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        fields = EVENT_FIELDS.copy()
        fields.pop(0)
        return dp.ListDisplay(fields=fields)

    def get_instance_display(self) -> Display:
        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["event_type", "person", "minute"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(800): Layout(
                            grid_template_areas=[
                                ["event_type", "."],
                                ["person", "minute"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                        ),
                        lte(450): Layout(
                            grid_template_areas=[
                                ["event_type"],
                                ["person"],
                                ["minute"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
            ]
        )


class EventTypeDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        return dp.ListDisplay(fields=EVENT_TYPE_FIELDS)

    def get_instance_display(self) -> Display:
        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["name", "sport"],
                                ["points", "icon"],
                                ["color", "color"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(650): Layout(
                            grid_template_areas=[
                                ["name"],
                                ["sport"],
                                ["points"],
                                ["icon"],
                                ["color"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
            ]
        )


class EventTypeSportDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        fields = EVENT_TYPE_FIELDS.copy()
        fields.pop(1)
        return dp.ListDisplay(fields=fields)

    def get_instance_display(self) -> Display:
        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["name", "points"],
                                ["icon", "color"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(650): Layout(
                            grid_template_areas=[
                                ["name"],
                                ["sport"],
                                ["points"],
                                ["icon"],
                                ["color"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
            ]
        )



from django_celery_beat.models import CrontabSchedule

from wbcore import viewsets

from .serializers import CrontabScheduleRepresentationSerializer


class CrontabScheduleRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = CrontabSchedule.objects.all()
    serializer_class = CrontabScheduleRepresentationSerializer



from wbcore.menus import ItemPermission, MenuItem

CURRENCY_MENUITEM = MenuItem(
    label="Currency",
    endpoint="wbcore:currency:currency-list",
    add=MenuItem(label="New Currency", endpoint="wbcore:currency:currency-list"),
    permission=ItemPermission(method=lambda request: request.user.is_staff),
)



# Generated by Django 5.0.14 on 2025-05-05 09:13

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('directory', '0012_alter_person_managers'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='clientmanagerrelationship',
            options={'permissions': (('administrate_clientmanagerrelationship', 'Can administrate Client Manager Relationship'),), 'verbose_name': 'Client Manager Relationship', 'verbose_name_plural': 'Client Manager Relationships'},
        ),
    ]



from django.conf import settings

from wbcore.serializers.fields import JSONField


class TranslationJSONField(JSONField):
    field_type = "translation_field"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.read_only = False
        self.required = False
        self.allow_null = True

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)
        representation["default_language"] = settings.LANGUAGE_CODE
        representation["languages"] = settings.MODELTRANS_AVAILABLE_LANGUAGES
        representation["fields"] = self.parent.Meta.model._meta.get_field("i18n").fields
        return key, representation



from dataclasses import dataclass, field

from django.utils.functional import cached_property
from django.utils.translation import gettext as _

from wbcore.metadata.configs.display.instance_display import Section
from wbcore.serializers import Serializer
from wbcore.serializers.fields import CharField


@dataclass
class ShareSite:
    serializers: list[Serializer] = field(default_factory=list)
    sections: list[Section] = field(default_factory=list)

    @cached_property
    def serializer_class(self) -> Serializer:
        return type(
            "ShareSerializer",
            tuple(self.serializers),
            dict(widget_endpoint=CharField(label=_("Widget URL"), read_only=True)),
        )

    @cached_property
    def sorted_sections(self) -> list[Section]:
        return list(map(lambda x: x["section"], sorted(self.sections, key=lambda x: x["weight"])))


share_site = ShareSite()



import json
from typing import Any

from wbcore.contrib.io.models import ImportSource

DUPLICATIONS_MAP = {"GBP": {"currency_key": "GBX", "factor": 100}}


def parse(import_source: ImportSource) -> dict[str, Any]:
    content = json.load(import_source.file)
    data = []
    for row in content:
        if row["success"]:
            val_date = row["date"]
            for currency_key, value in row["rates"].items():
                if duplicated_currency_data := DUPLICATIONS_MAP.get(currency_key, None):
                    data.append(
                        {
                            "currency__provider_id": duplicated_currency_data["currency_key"],
                            "currency__key": duplicated_currency_data["currency_key"],
                            "value": value * duplicated_currency_data["factor"],
                            "date": val_date,
                        }
                    )
                data.append(
                    {
                        "currency__provider_id": currency_key,
                        "currency__key": currency_key,
                        "value": value,
                        "date": val_date,
                    }
                )

    return {"data": data}



import urllib
from contextlib import suppress

from django.contrib.auth import get_user_model
from django.core.validators import URLValidator, ValidationError
from django.db import models
from django.urls import Resolver404, resolve
from django.utils.functional import cached_property

from wbcore.contrib.notifications.utils import base_domain, create_notification_type


class Notification(models.Model):
    title = models.CharField(max_length=255)
    body = models.TextField(null=True, blank=True)

    user = models.ForeignKey(to=get_user_model(), related_name="notifications_notifications", on_delete=models.CASCADE)
    notification_type = models.ForeignKey(
        to="notifications.NotificationType", related_name="notifications", on_delete=models.CASCADE
    )
    endpoint = models.CharField(max_length=2048, null=True, blank=True)

    created = models.DateTimeField(auto_now_add=True)
    sent = models.DateTimeField(null=True, blank=True)
    read = models.DateTimeField(null=True, blank=True)

    def __str__(self) -> str:
        return f"{self.user} {self.title}"

    class Meta:
        verbose_name = "Notification"
        verbose_name_plural = "Notifications"

        notification_types = [
            create_notification_type(
                "workbench.system",
                "System Notifications",
                "System Notifications.",
                True,
                True,
                False,
            ),
        ]

    @cached_property
    def is_endpoint_internal(self) -> bool:
        with suppress(Resolver404):
            if self.endpoint:
                resolve(
                    urllib.parse.urlsplit(urllib.parse.unquote(self.endpoint)).path
                )  # we need to truncate query parameters
                return True
        return False

    @cached_property
    def is_endpoint_valid(self) -> bool:
        try:
            URLValidator()(self.endpoint)
            return True
        except ValidationError:
            return False

    def get_full_endpoint(self, as_shareable_internal_link: bool = False) -> str | None:
        if self.is_endpoint_internal:
            if as_shareable_internal_link:
                return f"{base_domain()}?widget_endpoint={self.endpoint}"
            else:
                return f"{base_domain()}{self.endpoint}"
        elif self.is_endpoint_valid:
            return self.endpoint

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcore:notifications:notification"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{title}}"



# Generated by Django 5.0.12 on 2025-02-18 09:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('io', '0007_alter_exportsource_query_params'),
    ]

    operations = [
        migrations.AddField(
            model_name='importsource',
            name='resource_kwargs',
            field=models.JSONField(blank=True, default=dict),
        ),
    ]



from rest_framework.reverse import reverse

from wbcore import serializers
from wbcore.contrib.example_app.models import Sport


class SportModelSerializer(serializers.ModelSerializer):
    @serializers.register_only_instance_resource()
    def leagues_inline(self, instance: Sport, request, user, view, **kwargs) -> dict[str, str]:
        if not self.context.get("request"):
            return {}
        return {
            "existing_leagues": reverse(
                "example_app:league-sport-list", args=[instance.pk], request=self.context["request"]
            )
        }

    @serializers.register_only_instance_resource()
    def event_types_inline(self, instance: Sport, request, user, view, **kwargs) -> dict[str, str]:
        if not self.context.get("request"):
            return {}
        return {
            "associated_event_types": reverse(
                "example_app:eventtype-sport-list", args=[instance.pk], request=self.context["request"]
            )
        }

    class Meta:
        model = Sport
        fields = "__all__"


class SportRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="example_app:sport-detail")

    class Meta:
        model = Sport
        fields = ("id", "name", "_detail")



from .person import (
    SportPersonModelViewSet,
    SportPersonRepresentationViewSet,
    PlayerModelViewSet,
    PlayerTeamModelViewSet,
    PlayerRepresentationViewSet,
    TreeViewPlayerModelViewSet,
    SportPersonToolTipViewset,
)
from .league import LeagueModelViewSet, LeagueRepresentationViewSet, LeagueSportModelViewSet
from .stadium import StadiumModelViewSet, StadiumRepresentationViewSet
from .team import TeamModelViewSet, TeamRepresentationViewSet, TeamStadiumModelViewSet
from .sport import SportModelViewSet, SportRepresentationViewSet
from .role import RoleModelViewSet, RoleRepresentationViewSet
from .league import LeagueModelViewSet
from .match import MatchModelViewSet, MatchRepresentationViewSet, MatchStadiumModelViewSet, MatchLeagueModelViewSet
from .event import (
    EventModelViewSet,
    EventRepresentationViewSet,
    EventTypeRepresentationViewSet,
    EventTypeModelViewSet,
    EventTypeSportModelViewSet,
    EventMatchModelViewSet,
    PlayerStatisticsChartModelViewSet,
    LeaguePlayerStatisticsModelViewSet,
    LeagueTeamStatisticsModelViewSet,
)
from .teamresult import TeamResultsModelViewSet, TeamResultsRepresentationViewSet, TeamResultsLeagueModelViewSet
from .season import SeasonModelViewSet, SeasonRepresentationViewSet



from typing import Dict, Generator, Iterable, Type


def get_inheriting_subclasses(
    _class, only_leaf: bool = True, only_managed_model_classes: bool = True
) -> Generator[Type, None, None]:
    """
    Utility function to yield all inheriting subclasses. Useful to get all children classes for instance.

    Args:
        _class: The class to yield the subclasses from
        only_leaf: If True, returns only the leaf nodes (Default: True)
        only_managed_model_classes: If True, returns only the class implementing the managed Django model (Default: True)
    Returns:
        Generator of classes
    """
    subclasses = _class.__subclasses__()
    for subclass in subclasses:
        yield from get_inheriting_subclasses(subclass, True, only_managed_model_classes)
    if (not only_leaf or len(_class.__subclasses__()) == 0) and (
        not only_managed_model_classes
        or (
            only_managed_model_classes
            and (_meta := getattr(_class, "_meta", None))
            and _meta.managed
            and hasattr(_class, "objects")
        )
    ):
        yield _class


def uniquify_dict_iterable(iterable: Iterable[Dict], unique_key: str) -> Iterable[Dict]:
    keys = list()
    for item in iterable:
        if key := item.get(unique_key):
            if key not in keys:
                keys.append(key)
                yield item
        else:
            yield item



from typing import Optional

from django.utils.translation import gettext as _

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class DocumentModelRelationshipViewConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="content_type", label=_("Type")),
                dp.Field(key="content_object_repr", label=_("Instance"), width=500),
            ],
        )

    def get_instance_display(self) -> dp.Display:
        return dp.create_simple_display([["content_type"], ["object_id"]])



from selenium.webdriver.remote.webdriver import WebDriver

from wbcore.contrib.directory.factories import (
    ClientManagerRelationshipFactory,
    CompanyFactory,
    CompanyTypeFactory,
    CustomerStatusFactory,
    EmployerEmployeeRelationshipFactory,
    EntryFactory,
    PersonFactory,
    SpecializationFactory,
)
from wbcore.contrib.directory.models import (
    ClientManagerRelationship,
    Company,
    CustomerStatus,
    Person,
)
from wbcore.contrib.directory.serializers import (
    ClientManagerModelSerializer,
    CompanyModelSerializer,
    PersonModelSerializer,
)
from wbcore.test import (
    click_element_by_path,
    fill_out_form_fields,
    open_create_instance,
)

# from wbcrm.factories import ActivityFactory


def set_up_companies() -> list[Company]:
    """Creates three companies (called Company A to C), for testing purposes.

    In addition to the three companies, two company types and two activities are created in which the companies participate.\n
    Company A and Company C are of company type "Type A", Company B is of the type "Type B".\n
    Company A participates in an Activity that was 15 days ago, Company B and C participate in an Activity that was 60 days ago.
    """
    type_a = CompanyTypeFactory(title="Type A")
    type_b = CompanyTypeFactory(title="Type B")
    com_a = CompanyFactory(name="Company A", type=type_b)
    com_b = CompanyFactory(name="Company B", type=type_a)
    com_c = CompanyFactory(name="Company C", type=type_b)
    return [com_a, com_b, com_c]


def set_up_persons() -> list[Person]:
    """Creates three persons (called Henry Kalb, Konrad Zuse & Ada Lovelace), for testing purposes."""
    status_a = CustomerStatusFactory(title="Status A")
    status_b = CustomerStatusFactory(title="Status B")
    status_c = CustomerStatusFactory(title="Status C")
    company_a = CompanyFactory(customer_status=status_a)
    company_b = CompanyFactory(customer_status=status_b)
    company_c = CompanyFactory(customer_status=status_c)
    spec_a = SpecializationFactory(title="Specialization A")
    spec_b = SpecializationFactory(title="Specialization B")
    spec_c = SpecializationFactory(title="Specialization C")
    person_a = PersonFactory(first_name="Henry", last_name="Kalb", prefix=Person.Prefix.MR)
    person_b = PersonFactory(first_name="Konrad", last_name="Zuse", prefix=Person.Prefix.DR)
    person_c = PersonFactory(first_name="Ada", last_name="Lovelace", prefix=Person.Prefix.MRS)
    person_a.specializations.set([spec_a])
    person_b.specializations.set([spec_b, spec_c])
    person_c.specializations.set([spec_c])
    EmployerEmployeeRelationshipFactory(employer=company_a, employee=person_a, primary=True)
    EmployerEmployeeRelationshipFactory(employer=company_a, employee=person_b, primary=False)
    EmployerEmployeeRelationshipFactory(employer=company_b, employee=person_b, primary=True)
    EmployerEmployeeRelationshipFactory(employer=company_c, employee=person_c, primary=True)
    return [person_a, person_b, person_c]


def create_new_company_instance(
    driver: WebDriver,
    field_list: list[str],
    name="Test Company",
    status_title="Test Status",
    is_create_instance_open=True,
) -> Company:
    """A function that automatically creates a new company for selenium e2e-tests. After creating the instance this function will close the create-widget.

    Args:
        driver (WebDriver): The Selenium webdriver.
        field_list (list[str]): List of fields to be filled in the creation mask. The field names must match the names in the CompanyModelSerializer.
        name (str, optional): The title for the new company instance. Defaults to "Test Company".
        status_title (str, optional): The status for the new company instance. Defaults to "Test Status".
        is_create_instance_open (bool, optional): Should be true if the create-widget is already open. Defaults to True.

    Returns:
        Company: The newly created company.
    """
    if not is_create_instance_open:
        open_create_instance(driver, "CRM", "Create Company")

    if CustomerStatus.objects.filter(title=status_title).exists():
        status = CustomerStatus.objects.get(title=status_title)
    else:
        status = CustomerStatusFactory(title=status_title)
    company: Company = CompanyFactory.build(name=name, customer_status=status)
    serializer = CompanyModelSerializer(company)

    fill_out_form_fields(driver, serializer, field_list, company)
    click_element_by_path(driver, "//button[@label='Save and close']")
    return company


def create_new_person_instance(
    driver: WebDriver,
    field_list: list[str],
    first_name: str,
    last_name: str,
    prefix: str,
    is_create_instance_open=True,
) -> Person:
    """A function that automatically creates a new person for selenium e2e-tests. After creating the instance this function will close the create-widget.

    Args:
        driver (WebDriver): The Selenium webdriver.
        field_list (list[str]): List of fields to be filled in the creation mask. The field names must match the names in the PersonModelSerializer.
        first_name (str): The first name for the newly created person.
        last_name (str):  The last name for the newly created person.
        prefix (str):  The prefix for the newly created person.
        is_create_instance_open (bool, optional): Should be true if the create-widget is already open. Defaults to True.

    Returns:
        Person: The newly created person.
    """
    if not is_create_instance_open:
        open_create_instance(driver, "CRM", "Create Person")

    person: Person = PersonFactory.build(first_name=first_name, last_name=last_name, prefix=prefix)
    serializer = PersonModelSerializer(person)

    fill_out_form_fields(driver, serializer, field_list, person)
    click_element_by_path(driver, "//button[@label='Save and close']")
    return person


def create_new_cmr_instance(
    driver: WebDriver,
    field_list: list[str],
    is_primary=True,
    is_create_instance_open=True,
) -> ClientManagerRelationship:
    """A function that automatically creates a new client-manager-relationship for selenium e2e-tests. After creating the instance this function will close the create-widget.

    Args:
        driver (WebDriver): The Selenium webdriver.
        field_list (list[str]): List of fields to be filled in the creation mask. The field names must match the names in the ClientManagerModelSerializer.
        is_primary (str): True if the relationship shell be a primary relationship. Defaults to True.
        is_create_instance_open (bool, optional): Should be true if the create-widget is already open. Defaults to True.

    Returns:
        ClientManagerRelationship: The newly created CMR.
    """
    if not is_create_instance_open:
        open_create_instance(driver, "CRM", "Create Client Manager Relationship")
    manager = PersonFactory()
    entry = EntryFactory()
    cmr = ClientManagerRelationshipFactory.build(client=entry, relationship_manager=manager, primary=is_primary)
    serializer = ClientManagerModelSerializer(cmr)
    fill_out_form_fields(driver, serializer, field_list, cmr)
    click_element_by_path(driver, "//button[@label='Save and close']")
    return cmr



from wbcore.contrib.documents.models import Document


class DocumentMixin:
    @property
    def documents(self):
        return Document.get_for_object(self)

    def attach_document(self, document):
        document.link(self)



import pytest
from selenium.webdriver.common.action_chains import ActionChains

from wbcore.contrib.authentication.factories import SuperUserFactory
from wbcore.contrib.authentication.models import User
from wbcore.contrib.example_app.models import SportPerson
from wbcore.contrib.example_app.serializers import SportPersonModelSerializer
from wbcore.contrib.example_app.tests.e2e import create_new_person_instance
from wbcore.test import (
    click_element_by_path,
    click_new_button,
    delete_list_entry,
    edit_list_instance,
    is_counter_as_expected,
    is_string_not_visible,
    is_text_visible,
    open_menu_item,
    set_up,
)

USER_PASSWORD = "User_Password"


@pytest.mark.skip(reason="no way of currently testing this")
@pytest.mark.django_db
class TestPerson:
    def test_create_edit_delete_person(self, live_server, selenium):
        # Creating a test user and login to the WB
        user: User = SuperUserFactory(plaintext_password=USER_PASSWORD)
        actions = ActionChains(selenium, 1000)
        set_up(selenium, live_server, user.email, USER_PASSWORD)

        # -----> CREATE A NEW PERSON <----- #
        person_a = create_new_person_instance(selenium, ["first_name", "last_name"], False)
        open_menu_item(selenium, "Persons")
        # TODO find a solution how to handle the case, that a navigation element has the same name as a subelement.
        click_element_by_path(selenium, '//div[@title="Persons"]')
        click_element_by_path(selenium, '//span[@title="Create Person"]//parent::div//div')

        window_width = (selenium.get_window_size().get("width")) / 1.75
        actions.move_by_offset(window_width, 0).perform()
        actions.click()

        assert is_counter_as_expected(selenium, SportPerson.objects.count())
        assert is_text_visible(selenium, person_a.last_name)

        click_new_button(selenium)
        assert is_text_visible(selenium, "Create Person")
        person_b = create_new_person_instance(selenium, ["first_name", "last_name"])

        assert is_counter_as_expected(selenium, SportPerson.objects.count())
        assert is_text_visible(selenium, person_b.last_name)

        # ------> EDIT A NEW PERSON <------ #

        edit_list_instance(
            selenium, actions, person_a.last_name, SportPersonModelSerializer(person_a), {"last_name": "Foobar"}
        )
        assert is_string_not_visible(selenium, person_a.last_name)
        assert is_text_visible(selenium, "Foobar")
        assert is_counter_as_expected(selenium, SportPerson.objects.count())

        # -----> Delete <----- #

        delete_list_entry(selenium, actions, person_b.last_name)
        assert is_counter_as_expected(selenium, SportPerson.objects.count())
        assert is_text_visible(selenium, "Foobar")
        assert is_string_not_visible(selenium, person_b.last_name)



import factory

from wbcore.contrib.example_app.models import Sport


class SportFactory(factory.django.DjangoModelFactory):
    name = factory.Sequence(lambda n: "Sport %d" % n)

    rules = factory.Faker("text", max_nb_chars=100)
    match_duration = factory.Faker("pyint", min_value=1, max_value=250)

    class Meta:
        model = Sport



from django.utils.translation import gettext as _

from wbcore.metadata.configs.titles import TitleViewConfig


class SportTitleConfig(TitleViewConfig):
    def get_list_title(self) -> str:
        return _("Sports")

    def get_create_title(self) -> str:
        return _("Create Sport")

    def get_instance_title(self) -> str:
        return _("Sport")



from django.db.models import Case, Value, When
from django.db.models.query import F, QuerySet
from rest_framework import status
from rest_framework.request import Request
from rest_framework.response import Response

from wbcore import viewsets
from wbcore.contrib.notifications.models import (
    NotificationType,
    NotificationTypeSetting,
)
from wbcore.contrib.notifications.serializers import (
    NotificationTypeRepresentationSerializer,
    NotificationTypeSettingModelSerializer,
)

from ...icons import WBIcon
from .configs.notification_types import NotificationTypeSettingDisplayConfig, NotificationTypeSettingEndpointConfig


class NotificationTypeRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = NotificationType.objects.all()
    serializer_class = NotificationTypeRepresentationSerializer


class NotificationTypeSettingModelViewSet(viewsets.ModelViewSet):
    queryset = NotificationTypeSetting.objects.all()
    serializer_class = NotificationTypeSettingModelSerializer

    endpoint_config_class = NotificationTypeSettingEndpointConfig
    display_config_class = NotificationTypeSettingDisplayConfig

    search_fields = ["notification_type__title", "notification_type__help_text"]

    def create(self, request: Request) -> Response:
        return Response(status=status.HTTP_403_FORBIDDEN)

    def destroy(self, request: Request, pk=None) -> Response:
        return Response(status=status.HTTP_403_FORBIDDEN)

    def get_queryset(self) -> QuerySet[NotificationTypeSetting]:
        return (
            super()
            .get_queryset()
            .filter(user=self.request.user)
            .annotate(
                help_text=F("notification_type__help_text"),
                locked=F("notification_type__is_lock"),
                locked_icon=Case(When(locked=True, then=Value(WBIcon.LOCK.icon)), default=Value(None)),
            )
        )
